from __future__ import annotations
from cetpy.Modules.SysML._ValueProperty import ValueProperty, value_property, \
    DeterminationTest, AggregateValueProperty, ValuePrinter
from cetpy.Modules.SysML._PartProperty import PartProperty
from cetpy.Modules.SysML._FlowProperty import FlowProperty
from cetpy.Modules.SysML._Block import Block
from cetpy.Modules.SysML._Port import Port
from cetpy.Modules.SysML._ContinuousPort import ContinuousPort
